"""Class that uses the AQUA internal data model"""
import xarray as xr
from aqua.core.data_model import CoordTransformer
from aqua.core.logger import log_history, log_configure

DEFAULT_DATAMODEL = "aqua"

class FixerDataModel:
    """
    Class that uses the AQUA internal data model to apply to the data
    Further fixes to coordinates and dimensions can be applied.

    Args:
        fixes (dict): Dictionary containing the fixes to be applied.
        loglevel (int, optional): Log level for logging. Defaults to None.
    """ 

    def __init__(self, fixes=None, loglevel=None):

        self.fixes = fixes
        self.logger = log_configure(log_level=loglevel, log_name='FixerDataModel')
        self.loglevel = loglevel

    def apply_datamodel(self, data: xr.Dataset):
        """
        Apply fixes to the data model

        Arguments:
            data (xr.Dataset):  input dataset to process

        Returns:
            The processed input dataset
        """
        if self.fixes is None:
            return data
        
        datamodel = self.fixes.get("data_model", DEFAULT_DATAMODEL)
        if datamodel:
            data = CoordTransformer(data, loglevel=self.loglevel).transform_coords(name=datamodel)

        # Extra coordinate handling
        data = self._fix_dims(data)
        data = self._fix_coord(data)

        return data


    def fix_area(self, area: xr.DataArray):
        """
        Apply fixes to the area file

        Arguments:
            area (xr.DataArray):  area file to be fixed

        Returns:
            The fixed area file (xr.DataArray)
        """
        if self.fixes is None:  # No fixes available
            return area
        else:
            self.logger.debug("Applying fixes to area file")
            # This operation is a duplicate, rationalization with fixer method is needed
            #src_datamodel = self.fixes_dictionary["defaults"].get("src_datamodel", None)
            #src_datamodel = self.fixes.get("data_model", src_datamodel)

            #if src_datamodel:
            #    area = self.change_coord_datamodel(area, src_datamodel, self.dst_datamodel)
            area = CoordTransformer(area, loglevel=self.loglevel).transform_coords()

            return area

    def _fix_coord(self, data: xr.Dataset):
        """
        Other than the data_model we can apply other fixes to the coordinates
        reading them from the fixes file, in the coords section.
        Units override can also be specified.

        Arguments:
            data (xr.Dataset):  input dataset to process

        Returns:
            The processed input dataset
        """
        if self.fixes is None:
            return data

        coords_fix = self.fixes.get("coords", None)

        if coords_fix:
            coords = list(coords_fix.keys())
            self.logger.debug("Coordinates to be checked: %s", coords)

            for coord in coords:
                src_coord = coords_fix[coord].get("source", None)
                tgt_units = coords_fix[coord].get("tgt_units", None)

                if src_coord:
                    if src_coord in data.coords:
                        data = data.rename({src_coord: coord})
                        self.logger.debug("Coordinate %s renamed to %s", src_coord, coord)
                        log_history(data[coord], f"Coordinate {src_coord} renamed to {coord} by fixer")
                    else:
                        self.logger.warning("Coordinate %s not found", src_coord)

                if tgt_units:
                    if coord in data.coords:
                        self.logger.debug("Coordinate %s units set to %s", coord, tgt_units)
                        self.logger.debug("Please notice that this is an override, no unit conversion has been applied")
                        data[coord].attrs['units'] = tgt_units
                        log_history(data[coord], f"Coordinate {coord} units set to {tgt_units} by fixer")
                    else:
                        self.logger.warning("Coordinate %s not found", coord)

        return data
    
    def _fix_dims(self, data: xr.Dataset):
        """
        Other than the data_model we can apply other fixes to the dimensions
        reading them from the fixes file, in the dims section.

        Arguments:
            data (xr.Dataset):  input dataset to process

        Returns:
            The processed input dataset
        """
        if self.fixes is None:
            return data

        dims_fix = self.fixes.get("dims", None)

        if dims_fix:
            dims = list(dims_fix.keys())
            self.logger.debug("Dimensions to be checked: %s", dims)

            for dim in dims:
                src_dim = dims_fix[dim].get("source", None)

                if src_dim and src_dim in data.dims:
                    data = data.rename_dims({src_dim: dim})
                    self.logger.debug("Dimension %s renamed to %s", src_dim, dim)
                    log_history(data, f"Dimension {src_dim} renamed to {dim} by fixer")
                else:
                    self.logger.warning("Dimension %s not found", dim)

        return data



import pytest
from aqua import Reader
import pandas as pd
from conftest import APPROX_REL, LOGLEVEL

approx_rel = APPROX_REL
loglevel = LOGLEVEL

@pytest.fixture
def reader_instance():
    return Reader(model="IFS", exp="test-tco79", source="intake-esm-test",
                  areas=False, fix=False, loglevel=loglevel)

# aqua class for tests
@pytest.mark.aqua
class TestAqua:
    """ESM tests for AQUA reader"""

    def test_retrieve_esm_shape(self, reader_instance):
        """
        Test if the retrieve method returns data with the expected shape
        """
        data = reader_instance.retrieve()
        assert len(data) > 0
        assert data['2t'].shape == (2, 28480)

    def test_retrieve_esm_value(self, reader_instance):
        """
        Test if the retrieve method returns data with the expected average value
        """
        data = reader_instance.retrieve()
        assert data["2t"].mean().values == pytest.approx(286.48692342, rel=approx_rel)

    def test_retrieve_esm_var(self, reader_instance):
        """
        Test if the retrieve method returns data with the expected average value
        if a variable is specified
        """
        data = reader_instance.retrieve(var='2t')
        assert data["2t"].mean().values == pytest.approx(286.48692342, rel=approx_rel)

    def test_retrieve_esm_date_slice(self, reader_instance):
        """
        Test if the retrieve method correctly slices data by startdate and enddate
        """
        data = reader_instance.retrieve(startdate='2020-01-20', enddate='2020-01-20')
        assert len(data) > 0
        assert '2t' in data
        assert 'time' in data.dims
        # Verify the time range is correct
        assert data.time.min().dt.date.values == pytest.approx(pd.Timestamp('2020-01-20').date())
        assert data.time.max().dt.date.values == pytest.approx(pd.Timestamp('2020-01-20').date())



"""YAML utility functions"""

import os
from string import Template as DefaultTemplate
from jinja2 import Template
from collections import defaultdict
from ruamel.yaml import YAML
from tempfile import TemporaryDirectory
from aqua.core.logger import log_configure
import yaml  # This is needed to allow YAML override in intake


def construct_yaml_merge(loader, node):
    """
    This function is used to enable override in yaml for intake
    """
    if isinstance(node, yaml.ScalarNode):
        # Handle scalar nodes
        return loader.construct_scalar(node)
    else:
        # Handle sequence nodes
        maps = []
        for subnode in node.value:
            maps.append(loader.construct_object(subnode))
        result = {}
        for dictionary in reversed(maps):
            result.update(dictionary)
        return result


# Run this to enable YAML override for the yaml package when using SafeLoader in intake 
yaml.SafeLoader.add_constructor(
            'tag:yaml.org,2002:merge',
            construct_yaml_merge)


def load_multi_yaml(folder_path: str | None = None, filenames: list | None = None,
                    definitions: str | dict | None = None, **kwargs):
    """
    Load and merge yaml files.
    If a filenames list of strings is provided, only the yaml files with
    the matching full path will be merged.
    If a folder_path is provided, all the yaml files in the folder will be merged.

    Args:
        folder_path (str, optional): the path of the folder containing the yaml
                                        files to be merged.
        filenames (list, optional): the list of the yaml files to be merged.
        definitions (str or dict, optional): name of the section containing string template
                                                definitions or a dictionary with the same

    Keyword Args:
        loglevel (str, optional): the loglevel to be used, default is 'WARNING'

    Returns:
        A dictionary containing the merged contents of all the yaml files.
    """
    yaml = YAML()  # default, if not specified, is 'rt' (round-trip) # noqa F841

    if isinstance(definitions, str):  # if definitions is a string we need to read twice
        yaml_dict = _load_merge(folder_path=folder_path, definitions=None,
                                filenames=filenames, **kwargs)  # Read without definitions
        definitions = yaml_dict.get(definitions)
        yaml_dict = _load_merge(folder_path=folder_path, definitions=definitions,
                                filenames=filenames, **kwargs)  # Read again with definitions
    else:  # if a dictionary or None has been passed for definitions we read only once
        yaml_dict = _load_merge(folder_path=folder_path, definitions=definitions,
                                filenames=filenames, **kwargs)

    return yaml_dict


def load_yaml(infile: str, definitions: str | dict | None = None, jinja: bool = True):
    """
    Load yaml file with template substitution

    Args:
        infile (str): a file path to the yaml
        definitions (str or dict, optional): name of the section containing string template
                                             definitions or a dictionary with the same content
        jinja: (bool): jinja2 templating is used instead of standard python templating. Default is true.
    Returns:
        A dictionary with the yaml file keys
    """

    if not os.path.exists(infile):
        raise ValueError(f'ERROR: {infile} not found: you need to have this configuration file!')

    yaml = YAML(typ='rt')  # default, if not specified, is 'rt' (round-trip)

    cfg = None
    # Load the YAML file as a text string
    with open(infile, 'r', encoding='utf-8') as file:
        yaml_text = file.read()

    if isinstance(definitions, str):  # if it is a string extract from original yaml, else it is directly a dict
        cfg = yaml.load(yaml_text)
        definitions = cfg.get(definitions)

    if definitions:
        # perform template substitution with jinja
        if jinja:
            template = Template(yaml_text)
            rendered_yaml = template.render(definitions)
            cfg = yaml.load(rendered_yaml)
        # use default python templating
        else:
            template = DefaultTemplate(yaml_text).safe_substitute(definitions)
            cfg = yaml.load(template)
    else:
        if not cfg:  # did we already load it ?
            cfg = yaml.load(yaml_text)

    return cfg


def dump_yaml(outfile=None, cfg=None, typ='rt'):
    """
    Dump to a custom yaml file

    Args:
        outfile(str):   a file path
        cfg(dict):      a dictionary to be dumped
        typ(str):       the type of YAML initialisation.
                        Default is 'rt' (round-trip)
    """
    # Initialize YAML object
    yaml = YAML(typ=typ)

    yaml.representer.add_representer(
        type(None),
        lambda self, _: self.represent_scalar('tag:yaml.org,2002:null', 'null')
    )

    # Check input
    if outfile is None:
        raise ValueError('ERROR: outfile not defined')
    if cfg is None:
        raise ValueError('ERROR: cfg not defined')

    # Dump the dictionary with a safe temporary directory
    # to avoid intake reading a partially written file

    # Ensure parent directory exists
    dest_dir = os.path.dirname(os.path.abspath(outfile))
    if dest_dir:  # Handle edge case where filename has no directory component
        os.makedirs(dest_dir, exist_ok=True)
    else:
        dest_dir = '.'  # Use current directory
    with TemporaryDirectory(dir=dest_dir) as tmpdirname:
        tmp_file = os.path.join(tmpdirname, "temp.yaml")
        with open(tmp_file, 'w', encoding='utf-8') as file:
            yaml.dump(cfg, file)
        os.replace(tmp_file, outfile)


def _load_merge(folder_path: str | None = None, filenames: list | None = None,
                definitions: str | dict | None = None, merged_dict: dict | None = None,
                loglevel: str = 'WARNING'):
    """
    Helper function for load_merge_yaml.
    Load and merge yaml files located in a given folder
    or a list of yaml files into a dictionary.

    Args:
        folder_path (str, optional):         the path of the folder containing the yaml
                                             files to be merged.
        filenames (list, optional):          the list of the yaml files to be merged.
        definitions (str or dict, optional): name of the section containing string template
                                             definitions or a dictionary with the same content
        merged_dict (dict, optional):        the dictionary to be updated with the yaml files
        loglevel (str, optional):            the loglevel to be used, default is 'WARNING'

    Returns:
        A dictionary containing the merged contents of all the yaml files.

    Raises:
        ValueError: if both folder_path and filenames are None or if both are not None.
    """
    logger = log_configure(log_name='yaml', log_level=loglevel)

    if merged_dict is None:
        logger.debug('Creating a new dictionary')
        merged_dict = defaultdict(dict)
    else:
        logger.debug('Updating an existing dictionary')

    if filenames is None and folder_path is None:
        raise ValueError('ERROR: at least one between folder_path or filenames must be provided')

    if filenames:  # Merging a list of files
        logger.debug(f'Files to be merged: {filenames}')
        for filename in filenames:
            yaml_dict = load_yaml(filename, definitions)
            for key, value in yaml_dict.items():
                merged_dict[key].update(value)

    if folder_path:  # Merging all the files in a folder
        logger.debug(f'Folder to be merged: {folder_path}')
        if not os.path.exists(folder_path):
            raise ValueError(f'ERROR: {folder_path} not found: it is required to have this folder!')
        else:  # folder exists
            for filename in os.listdir(folder_path):
                if filename.endswith(('.yml', '.yaml')):
                    file_path = os.path.join(folder_path, filename)
                    yaml_dict = load_yaml(file_path, definitions)
                    for key, value in yaml_dict.items():
                        merged_dict[key].update(value)

    logger.debug('Dictionary updated')
    logger.debug(f'Keys: {merged_dict.keys()}')

    return merged_dict



"""Module to transform coordinates of an Xarray object."""

import os
import xarray as xr
from metpy.units import units
from aqua.core.logger import log_configure, log_history
from aqua.core.util import load_yaml
from aqua.core.configurer import ConfigPath
from .coordidentifier import CoordIdentifier
from pint.errors import DimensionalityError


# Function to get the conversion factor
def units_conversion_factor(from_unit_str, to_unit_str):
    """
    Get the conversion factor between two units.
    """
    from_unit = units(from_unit_str)
    to_unit = units(to_unit_str)
    try:
        return from_unit.to(to_unit).magnitude
    except DimensionalityError:
        return None



class CoordTransformer():
    """
    Class to transform coordinates of an Xarray object.
    It aims at transforming the coordinates provided by the user into
    a standard format.
    """

    def __init__(self, data, loglevel='WARNING'):
        """
        Constructor of the CoordTransator class.

        Args:
            data (xr.Dataset or xr.DataArray): Xarray Dataset or DataArray object.
            loglevel (str, optional): Log level. Defaults to 'WARNING'.
        """
        if not isinstance(data, (xr.Dataset, xr.DataArray)):
            raise TypeError("data must be an Xarray Dataset or DataArray object.")
        self.loglevel = loglevel
        self.logger = log_configure(self.loglevel, 'CoordTransformer')

        self.data = data

        self.src_coords = CoordIdentifier(data.coords, loglevel=loglevel).identify_coords()
        self.tgt_coords = None
        self.gridtype = self._info_grid(data.coords)
        self.logger.info("Grid type: %s", self.gridtype)

    def load_data_model(self, name: str = "aqua"):
        """
        Load the default data model from the aqua.yaml file.

        Args:
            name (str): An installed data_model into aqua config, i.e. a YAML file

        Returns:
            dict: Target coordinates dictionary.
            str: Name of the target data model.
        """

        data_model_dir = os.path.join(ConfigPath().get_config_dir(), "data_model")
        data_model_file = os.path.join(data_model_dir, f"{name}.yaml")
        if not os.path.exists(data_model_file):
            raise FileNotFoundError(f"Data model file {data_model_file} not found.")
        self.logger.info("Loading data model from %s", data_model_file)
        data_yaml = load_yaml(data_model_file)
        return data_yaml

    def _info_grid(self, coords):
        """
        Identify the grid type of the Xarray object.
        To be used to check if the axis direction has to be reversed. 
        Args: 
            coords (xr.Coordinates): Coordinates of the Xarray object.
        Returns:
            str: The grid type of the Xarray object. 
            It can be Regular, Curvilinear or Unstructured.
        """

        lonname = self.src_coords.get('latitude')
        latname = self.src_coords.get('longitude')
        if lonname is None or latname is None:
            return "Unknown"

        lat = coords[latname.get('name')]
        lon = coords[lonname.get('name')]
        if lon.ndim == 2 and lat.ndim == 2:
            return "Curvilinear"
        if lon.dims != lat.dims:
            return "Regular"
        return "Unstructured"
    
    def transform_coords(self, name="aqua"):
        """
        Transform the coordinates of the Xarray object.

        Args:
            tgt_coords (dict, optional): Target coordinates dictionary. Defaults to None.

        Returns:
            xr.Dataset or xr.DataArray: The transformed dataset or dataarray.
        """

        if not isinstance(name, str):
            raise TypeError("name must be a string.")

        # multiple safety check
        self.logger.info("Target data model: %s", name)
        data_yaml = self.load_data_model(name)
        self.tgt_coords = data_yaml.get('data_model')
        outname = f"{data_yaml.get('name')} v{str(data_yaml.get('version'))}"

        if not isinstance(self.tgt_coords, dict):
            raise TypeError("tgt_coords must be a dictionary.")
        
        data = self.data
        for coord in self.tgt_coords:
            if coord in self.src_coords and self.src_coords[coord]:
                tgt_coord = self.tgt_coords[coord]
                src_coord = self.src_coords[coord]
                self.logger.info("Analysing coordinate: %s", coord)
                #self.logger.info("Transforming coordinate %s to %s", src_coord, tgt_coord)
                data = self.rename_coordinate(data, src_coord, tgt_coord)
                data = self.reverse_coordinate(data, src_coord, tgt_coord)
                data = self.convert_units(data, src_coord, tgt_coord)
                data = self.assign_attributes(data, tgt_coord)
            else:
                self.logger.info("Coordinate %s not found in source coordinates.", coord)

        log_history(data, f"Coordinates fixed toward {outname} datamodel")

        return data
    
    def rename_coordinate(self, data, src_coord, tgt_coord):
        """
        Rename coordinate if necessary.

        Args:
            data (xr.Dataset or xr.DataArray): The Xarray object.
            src_coord (dict): Source coordinate dictionary.
            tgt_coord (dict): Target coordinate dictionary.
        Returns:
            xr.Dataset or xr.DataArray: The Xarray object with renamed coordinate.
        """
        if src_coord['name'] != tgt_coord['name']:
            original_coords = list(data.coords)
            self.logger.info("Renaming coordinate %s to %s",
                            src_coord['name'], tgt_coord['name'])
            data = data.rename({src_coord['name']: tgt_coord['name']})
            log_history(data,
                        f"Renamed coordinate {src_coord['name']} to {tgt_coord['name']} by datamodel")

            # Ensure the AQUA dependent index is preserved
            if f"idx_{src_coord['name']}" in original_coords:
                index_name = f"idx_{src_coord['name']}"
                new_index_name = f"idx_{tgt_coord['name']}"
                self.logger.info("Renaming index %s to %s", index_name, new_index_name)
                data = data.rename({index_name: new_index_name})

            # unclear if this is fundamental
            # if tgt_coord['name'] in data.dims:
            #   self.logger.info("Preserving original dimension %s and index.", src_coord['name'])
            #   data = data.swap_dims({tgt_coord['name']: src_coord['name']})
            #   data = data.set_index({src_coord['name']: tgt_coord['name']})
                
            #data = data.rename({src_coord['name']: tgt_coord['name']})
            tgt_coord['bounds'] = f"{tgt_coord['name']}_bnds"
            data = self._rename_bounds(data, src_coord, tgt_coord)
        return data
    
    def _rename_bounds(self, data, src_coord, tgt_coord):
        """
        Rename bounds if necessary.

        Args:
            data (xr.Dataset or xr.DataArray): The Xarray object.
            src_coord (dict): Source coordinate dictionary.
            tgt_coord (dict): Target coordinate dictionary.
        
        Returns:
            xr.Dataset or xr.DataArray: The Xarray object with renamed bounds.
        """
        if src_coord['bounds'] is not None:
            if src_coord['bounds'] in data:
                self.logger.info("Renaming bounds %s to %s",
                                src_coord['bounds'], tgt_coord['bounds'])
                data = data.rename({src_coord['bounds']: tgt_coord['bounds']})
                data[tgt_coord['name']].attrs['bounds'] = tgt_coord['bounds']
            else:
                self.logger.info("Bounds %s not found in data.", src_coord['bounds'])
        return data

    def reverse_coordinate(self, data, src_coord, tgt_coord):
        """
        Reverse coordinate if necessary.

        Args:
            data (xr.Dataset or xr.DataArray): The Xarray object.
            src_coord (dict): Source coordinate dictionary.
            tgt_coord (dict): Target coordinate dictionary.
        
        Returns:
            xr.Dataset or xr.DataArray: The Xarray object with possibly reversed coordinate.
        """
        if 'stored_direction' not in tgt_coord:
            return data
        if tgt_coord['stored_direction'] not in ["increasing", "decreasing"]:
            raise ValueError(f"tgt direction must be 'increasing' or 'decreasing', not  {tgt_coord['stored_direction']}")
        if src_coord['stored_direction'] not in ["increasing", "decreasing"]:
            self.logger.warning("src direction is not 'increasing' or 'decreasing', but %s. Disabling reverse!", src_coord['stored_direction'])
            return data
        if src_coord['stored_direction'] != tgt_coord['stored_direction']:
            if self.gridtype == "Regular":
                self.logger.info("Reversing coordinate %s from %s to %s",
                                tgt_coord['name'], src_coord['stored_direction'], tgt_coord['stored_direction'])
                data = data.isel({tgt_coord['name']: slice(None, None, -1)})
                # add an attribute for regridder evalution
                data[tgt_coord['name']].attrs['flipped'] = 1
                log_history(data, f"Reversed coordinate {tgt_coord['name']} from {src_coord['stored_direction']} to {tgt_coord['stored_direction']} by datamodel")
            else:
                self.logger.info("Cannot reverse coordinate %s. Grid type is %s.",
                                    tgt_coord['name'], self.gridtype)
        return data
    
    
    def convert_units(self, data, src_coord, tgt_coord):
        """
        Convert units of the coordinate.
        """
        if 'units' not in tgt_coord:
            self.logger.debug("%s unit not found in target data model. Disabling unit conversion.", tgt_coord['name'])
            return data
        if 'units' not in src_coord:
            self.logger.warning("%s unit not found source data model. Disabling unit conversion.", src_coord['name'])
            return data
        if 'units' not in data[tgt_coord['name']].attrs:
            self.logger.warning("%s unit not found in data. Disabling unit conversion.", tgt_coord['name'])
            return data
        if src_coord['units'] != tgt_coord['units']:
            self.logger.info("Converting units of coordinate %s from %s to %s",
                            src_coord['name'], src_coord['units'], tgt_coord['units'])
            factor = units_conversion_factor(src_coord['units'], tgt_coord['units'])
            if factor is None:
                self.logger.warning(
                    "Incompatible unit conversion for coordinate %s: %s -> %s. Skipping conversion.",
                    src_coord['name'], src_coord['units'], tgt_coord['units']
                )
                return data
            if factor != 0:
                self.logger.info("Conversion factor is: %s ", factor)
                data = data.assign_coords({tgt_coord['name']: data[tgt_coord['name']]*factor})
                tgt_coord['bounds'] = f"{tgt_coord['name']}_bnds"
                data = self._convert_bounds(data, src_coord, tgt_coord, factor)
                log_history(data,
                            f"Converted units of coordinate {src_coord['name']} from {src_coord['units']} to {tgt_coord['units']} by datamodel")
            data[tgt_coord['name']].attrs['units'] = tgt_coord['units']
        return data
    
    def _convert_bounds(self, data, src_coord, tgt_coord, factor):
        """
        Convert units bounds of the coordinate.
        """
        if 'bounds' not in tgt_coord:
            return data
        if tgt_coord['bounds'] in data:
            self.logger.info("Converting bounds of coordinate %s from %s to %s",
                            src_coord['name'], src_coord['units'], tgt_coord['units'])
            data[tgt_coord['bounds']] = data[tgt_coord['bounds']]*factor
            data[tgt_coord['bounds']].attrs['units'] = tgt_coord['units']
        return data

    def assign_attributes(self, data, tgt_coord):
        """
        Assign attributes to the coordinate.
        """
        for key, value in tgt_coord.items():
            if key not in['name', 'units', 'positive', 'stored_direction', 'bounds']:
                if key not in data.coords[tgt_coord['name']].attrs:
                    self.logger.debug("Adding attribute %s to coordinate %s", key, tgt_coord['name'])
                    data.coords[tgt_coord['name']].attrs[key] = value
        return data
    

def counter_reverse_coordinate(data):
    """
    Flip back latitude if necessary
    """

    for coord in data.coords:
        if 'flipped' in data.coords[coord].attrs:
            data = data.isel({coord: slice(None, None, -1)})
            del data.coords[coord].attrs['flipped']
    return data


"""Fixer mixin for the Reader class"""
import re
import numpy as np
from aqua.core.util import to_list, convert_units, get_eccodes_attr
from aqua.core.logger import log_history, log_configure
from .fixer_operator import FixerOperator
from .fixer_datamodel import FixerDataModel
from .evaluate_formula import EvaluateFormula

DEFAULT_DELTAT = 1

class Fixer():
    """Fixer module
    
    Args:
        fixer_name (str): The fixer name defined in the fixes dictionary
        datamodel (str): The target datamodel name
        fixes_dictionary (dict): The dictionary of fixes
        convention (name): The convention name
        metadata (dict): The metadata dictionary
        loglevel (str): The log level
    
    """

    def __init__(self, fixer_name=None, fixes_dictionary=None,
                 convention=None, metadata=None, loglevel='WARNING'):

        self.fixes_dictionary = fixes_dictionary
        self.fixer_name = fixer_name
        self.convention = convention
        self.metadata = metadata
        self.logger = log_configure(log_level=loglevel, log_name='Fixer')
        self.loglevel = loglevel
        self.fixes = self.find_fixes()
        self.deltat = self._define_deltat(default=DEFAULT_DELTAT)
        self.time_correction = False
        self.operator = FixerOperator(self.fixes, loglevel=loglevel)
        self.datamodel = FixerDataModel(self.fixes, loglevel=loglevel)


    def find_fixes(self):
        """
        Get the fixes for the required model, experiment and source.
        A convention dictionary is loaded and merged with the fixer_name.
        The default convention is eccodes-2.39.0.

        If not found, looks for default fixes for the model.
        Then source_fixes are loaded and merged with base_fixes.

        This second block of search is deprecated and will be removed in the future.

        Args:
            The fixer class

        Return:
            The fixer dictionary
        """
        # The convention dictionary is defined by stardard: eccodes and version: 2.39.0
        # At the actual stage 'eccodes' is the default and if not set to None or 'eccodes'
        # an error is raised in the Reader initialization
        convention_dictionary = self._load_convention_dictionary() if self.convention else None

        # Here we load the fixer_name, the specific fix that is merged with the convention dictionary
        # The merge is done only if the base_fixes dictionary has the 'convention' field matching the
        # convention dictionary.
        base_fixes = self._load_fixer_name()
        return self._combine_convention(base_fixes, convention_dictionary)

    def _load_convention_dictionary(self, version='2.39.0'):
        """
        Load the convention dictionary from the fixer file.
        The convention_name block should be called <convention>-<version>.

        Args:
            version: The convention name version. Default is 2.39.0

        Returns:
            The convention dictionary
        """
        convention_dictionary = self.fixes_dictionary.get("convention_name", None)
        if convention_dictionary is None:
            self.logger.error("Convention dictionary not found, will be disabled")
            return None

        convention_name = self.convention + '-' + version

        convention_dictionary = convention_dictionary.get(convention_name, None)
        if convention_dictionary is None:
            self.logger.error("No convention dictionary found for %s", convention_name)
            return None
        else:
            self.logger.info("Convention dictionary: %s", convention_name)

        return convention_dictionary

    def _combine_convention(self, base_fixes: dict, convention_dictionary: dict):
        """
        Combine convention dictionary with the fixes.
        It scan the 'vars' block and merge the convention dictionary with the fixer_name.
        If an item is new in the convention dictionary, it is added to the fixes.
        If the item is already present, non existing keys are added,
        otherwise the existing keys (in the base file) are kept.

        Args:
            base_fixes (dict): The base fixes (coming from the fixer_name)
            convention_dictionary (dict): The convention dictionary

        Returns:
            The final fixes, with the convention dictionary merged
        """
        if base_fixes is None:
            self.logger.info("No fixer_name found, only convention will be applied")
            base_fixes = {}
            base_convention = 'eccodes'
        elif convention_dictionary is None:
            self.logger.info("No convention dictionary found, only fixer_name will be applied")
            return base_fixes
        else:
            base_convention = base_fixes.get('convention', None)
        # We do not crash if the convention is not eccodes, but we log an error and return the base fixes
        convention = convention_dictionary.get('convention', None)
        if convention != 'eccodes':
            self.logger.error("Convention %s not supported, only eccodes is supported", convention)
            return base_fixes

        # We make sure that the convention is the same in the convention dictionary and in the fixer_name
        # Additionally, we check that the version is the same, knowing that for the moment we are not going to
        # document this feature since the version is hardcoded in the code.
        # TODO: version should be a parameter in the fixer_name and in the convention dictionary
        if base_convention is not None:
            if base_convention != convention and convention is not None:
                raise ValueError(f"The convention in the convention dictionary: {base_convention} is different from the fixer_name: {convention}")
            if 'version' in base_fixes and 'version' in convention_dictionary:
                if base_fixes['version'] != convention_dictionary['version']:
                    raise ValueError(f"The version in the convention dictionary: {base_fixes['version']} is different from the fixer_name: {convention_dictionary['version']}")
        else:
            self.logger.info("No convention found in the fixer_name, the convention dictionary will not be used")
            return base_fixes

        # Merge one by one the variables. This is done so that we can be careful at the level of the individual variables.
        # If a field for the specific variable is present in the base_fixes, it has priority.
        if 'vars' in convention_dictionary:
            if 'vars' not in base_fixes:
                base_fixes['vars'] = convention_dictionary['vars']
            else:  # A merge one variable by one is needed
                for var_key in convention_dictionary['vars'].keys():
                    if var_key in base_fixes['vars']:
                        # self.logger.debug("Variable %s already present in the fixes, merging...", var_key)
                        # This requires python >= 3.9
                        base_fixes['vars'][var_key] = convention_dictionary['vars'][var_key] | base_fixes['vars'][var_key]
                        # We need to check that only one between 'source' and 'derived' is present.
                        # If both are present, we give priority to 'derived' and log an info.
                        if 'source' in base_fixes['vars'][var_key] and 'derived' in base_fixes['vars'][var_key]:
                            self.logger.info("Variable %s has both 'source' and 'derived' in the fixes, 'derived' will be used",  # noqa: E501
                                             var_key)
                            base_fixes['vars'][var_key].pop('source')
                    else:
                        # This is a new variable to fix
                        # We need to manipulate the convention_dictionary to set the grib flag
                        # TODO: expand to other formats (cmor tables, etc.)
                        new_var = convention_dictionary['vars'][var_key]
                        if convention == 'eccodes':
                            new_var['grib'] = convention_dictionary['vars'][var_key].get('grib', True)
                        base_fixes['vars'][var_key] = new_var
        else:
            self.logger.warning("No 'vars' block found in the convention dictionary")

        return base_fixes

    def _load_fixer_name(self):
        """
        Load the fixer_name reading from the metadata of the catalog.
        If the fixer_name has a parent, load it and merge it giving priority to the child.
        """
        # if fixer name is found, get it
        if self.fixer_name is not None:
            self.logger.info('Fix names in metadata is %s', self.fixer_name)
            fixes = self.fixes_dictionary["fixer_name"].get(self.fixer_name)
            if fixes is None:
                self.logger.error("The requested fixer_name %s does not exist in fixes files", self.fixer_name)
                return None
            else:
                self.logger.info("Fix names %s found in fixes files", self.fixer_name)

                if 'parent' in fixes:
                    parent_fixes = self.fixes_dictionary["fixer_name"].get(fixes['parent'])
                    if parent_fixes is not None:
                        self.logger.info("Parent fix %s found! Mergin with fixer_name fixes %s!", fixes['parent'],
                                         self.fixer_name)
                        fixes = self._merge_fixes(parent_fixes, fixes)
                    else:
                        self.logger.error("Parent fix %s defined but not available in the fixes file.", fixes['parent'])

            return fixes

        # if not fixes found, return fixes None
        return None

    @staticmethod
    def _merge_fixes(base, specific):
        """
        Small function to merge fixes. Base fixes will be used as a default
        and specific fixes will replace where necessary. Dictionaries will be merged
        for variables, with priority for the specific ones.

        Args:
            base (dict): Base fixes
            specific (dict): Specific fixes

        Return:
            dict with merged fixes
        """
        final = base
        for item in specific.keys():
            if item == 'vars' and item in base:
                final[item] = {**base[item], **specific[item]}
            else:
                final[item] = specific[item]

        return final


    def fixer(self, data, destvar, apply_unit_fix=True):
        """
        Perform fixes (var name, units, coord name adjustments) of the input dataset.

        Arguments:
            data (xr.Dataset):      the input dataset
            destvar (list of str):  the name of the desired variables to be fixed, if None all available variables are fixed
            apply_unit_fix (bool):  if to perform immediately unit conversions (which requite a product or an addition).
                                    The fixer sets anyway an offset or a multiplicative factor in the data attributes.
                                    These can be applied also later with the method `apply_unit_fix`. (true)
        Returns:
            A xarray.Dataset containing the fixed data and target units, factors and offsets in variable attributes.
        """

        # OLD NAMING SCHEME
        # unit: name of 'units' attribute
        # src_units: name of fixer source units
        # newunits: name of fixer target units

        # NEW NAMING SCHEME
        # tgt_units: target unit
        # fixer_src_units: name of fixer source units
        # fixer_tgt_units: name of fixer target units

        # Add extra units (might be moved somewhere else, function is at the bottom of this file)

        # Fix GRIB attribute names. This removes "GRIB_" from the beginning
        # for var in data.data_vars:
        #    data[var].attrs = {key.split("GRIB_")[-1]: value for key, value in data[var].attrs.items()}

        # if there are no fixes defined, return
        if self.fixes is None:
            return data

        # Default input datamodel
        #src_datamodel = self.fixes_dictionary["defaults"].get("src_datamodel", None)
        #self.logger.debug("Default input datamodel: %s", src_datamodel)

      
        # Special case for monthly deltat
        if self.deltat == "monthly":
            self.logger.info('%s deltat found, we will estimate a correction based on number of days per month', self.deltat)
            self.deltat = 3600*24
            self.time_correction = data.time.dt.days_in_month

        jump = self.fixes.get("jump", None)  # if to correct for a monthly accumulation jump

        # Special feature to fix corrupted data in first step of each month
        nanfirst_startdate = self.fixes.get("nanfirst_startdate", None)
        nanfirst_enddate = self.fixes.get("nanfirst_enddate", None)

        fixd = {}  # variables dictionary for name change: only for source, done as {source: var}
        varlist = {}  # variable dictionary for name change
        vars_to_fix = self.fixes.get("vars", None)  # variables with available fixes

        # check which variables need to be fixed among the requested ones
        vars_to_fix = self._check_which_variables_to_fix(vars_to_fix, destvar)

        if vars_to_fix:
            for var in vars_to_fix:

                # Dictionary of fixes of the single var
                varfix = vars_to_fix[var]

                # Get grib attributes if requested and fix name
                # This can be expanded to other formats in the future
                grib = varfix.get("grib", None)
                # We make sure also of the case were an user saw a grib: True
                # and decided to build a grib: False instead of just not using
                # the block
                if grib is not None and grib is not False:
                    # grib: True means that we're just going to use the default grib attributes
                    # associated with the variable name var
                    if isinstance(grib, bool):
                        attributes, shortname = self._get_variables_grib_attributes(var)
                    # grib: paramid is an option, this means that the variable name may not correspond
                    # to the shortname that it can be found within the attributes
                    elif isinstance(grib, int):
                        attributes, shortname = self._get_variables_grib_attributes(f"var{grib}")
                    else:
                        raise ValueError("grib should be either a boolean or an integer")
                else:
                    attributes = {}
                    shortname = var

                # Get extra attributes from fixer, leave empty dict otherwise
                attributes.update(varfix.get("attributes", {}))

                # Define the list of name changes
                varlist[var] = shortname

                # 1. source case. We want to be able to work with a list of sources to scan
                source = to_list(varfix.get("source", None))
                # We want to process a list of sources
                if source:
                    match = list(set(source) & set(data.variables))
                    if match:
                        # Having more than a match should be a problem for a dataset, we do not raise an error
                        # but we warn the user
                        if len(match) > 1:
                            self.logger.error("Multiple matches found for variable %s: %s, the first one will be taken",
                                              var, match)
                        # Even if we have only a match, we make sure that source is a string
                        source = match[0]

                        # If a gribcode is the source match, convert it to shortname to access it
                        if str(source).isdigit():
                            self.logger.info('The source %s is a grib code, need to convert it', source)
                            source = get_eccodes_attr(f'var{source}', loglevel=self.loglevel)['shortName']

                        # Here we update the fixd dictionary with the source and the variable
                        # The rename is done as {source: var} and at the end of the function
                        # This because the source name could be used in the derived formula
                        fixd.update({f"{source}": f"{var}"})
                        if source != var:
                            # We keep in the history the original variable name only if it is different from the target
                            log_history(data[source], f"Variable renamed {var} from {source} by fixer")
                    else:  # if there is no match
                        # We do not know in advance if the source is available, so we loop over all the available in the final
                        # merge of the fixes
                        self.logger.debug('While fixing variable %s, no match found with sources %s', var, source)
                        continue

                # 2. derived case: let's compute the formula it and create the new variable
                formula = varfix.get("derived", None)
                if formula:
                    # If the formula is the same as the variable name, we raise an error
                    # Asking for a derived variable that is also a source variable is not allowed
                    # since it may lead to changing how the fixer based on the source variable is applied
                    if formula == var:
                        self.logger.error('Derived variable %s cannot have the same name as the source variable, skipping it',
                                          var)
                        continue
                    try:
                        source = shortname
                        data[source] = EvaluateFormula(data=data, formula=formula, short_name=shortname,
                                                       loglevel=self.loglevel).evaluate()
                        attributes.update({"derived": formula})
                        self.logger.debug("Derived %s from %s", var, formula)
                        log_history(data[source], f"Variable {var}, derived with {formula} by fixer")
                    except KeyError:
                        # The variable could not be computed, let's skip it
                        if destvar is not None:
                            # issue an error if you are asking that specific variable!
                            self.logger.error('Requested derived variable %s cannot be computed, is it available?', shortname)
                        else:
                            self.logger.info('%s is defined in the fixes but cannot be computed, is it available?',
                                             shortname)
                        continue

                # safe check debugging
                self.logger.debug('Name of fixer var: %s', var)
                self.logger.debug('Name of data source var: %s', source)
                self.logger.debug('Name of target var: %s', shortname)

                # fix source units
                data = self._override_src_units(data, varfix, var, source)

                # update attributes to the data but the units
                tgt_units = None
                if attributes:
                    for att, value in attributes.items():
                        # Already adjust all attributes but not yet units
                        if att == "units":
                            tgt_units = value
                        else:
                            data[source].attrs[att] = value

                tgt_units = self._override_tgt_units(tgt_units, varfix, var)

                if "units" not in data[source].attrs:  # Houston we have had a problem, no units!
                    self.logger.error('Variable %s has no units!', source)

                # adjust units
                if tgt_units:

                    if tgt_units.count('{'):  # WHAT IS THIS ABOUT?
                        tgt_units = self.fixes_dictionary["defaults"]["units"]["shortname"][tgt_units.replace('{',
                                                                                                              '').replace('}',
                                                                                                                          '')]
                    self.logger.info("%s: converting units %s --> %s", var, data[source].units, tgt_units)
                    if data[source].units != tgt_units:
                        log_history(data[source], f"Converting units of {var}: from {data[source].units} to {tgt_units}")
                    conversion_dictionary = convert_units(data[source].units, tgt_units, deltat=self.deltat,
                                                          var=var, loglevel=self.loglevel)

                    # if some unit conversion is defined, modify the attributes and history for later usage
                    if conversion_dictionary:
                        data[source].attrs.update({"tgt_units": tgt_units})
                        for key, value in conversion_dictionary.items():
                            data[source].attrs.update({key: value})
                            self.logger.debug("Fixing %s to %s. Unit fix: %s=%f", source, var, key, float(value))
                            log_history(data[source], f"Fixing {source} to {var}. Unit fix: {key}={value}")
                    elif conversion_dictionary == {} and data[source].units != tgt_units:
                        self.logger.info("No conversion needed for %s, but units are renamed from %s to %s",
                                         var, data[source].units, tgt_units)
                        data[source].attrs.update({"units": tgt_units})

                # Set to NaN before a certain date
                mindate = varfix.get("mindate", None)
                if mindate:
                    data[source] = data[source].where(data.time >= np.datetime64(str(mindate)), np.nan)
                    data[source].attrs.update({"mindate": mindate})
                    self.logger.debug("Steps before %s set to NaN for variable %s", str(mindate), var)

        # Only now rename everything
        for item, value in fixd.items():
            if not data[item].attrs.get("derived", None):
                data = data.rename({item: value})
            else:
                self.logger.info("Variable %s is derived, it will not be renamed to %s", item, value)

        # decumulate if necessary and fix first of month if necessary
        if vars_to_fix:
            data = self.operator.wrapper_decumulate(data, self.deltat, vars_to_fix, varlist, jump)
            if nanfirst_enddate:  # This is a temporary fix for IFS data, run ony if an end date is specified
                data = self.operator.wrapper_nanfirst(data, vars_to_fix, varlist,
                                              startdate=nanfirst_startdate,
                                              enddate=nanfirst_enddate)

        if apply_unit_fix:
            for var in data.data_vars:
                self.operator.apply_unit_fix(data[var], time_correction=self.time_correction)

        # apply time shift if necessary
        data = self.operator.timeshifter(data)

        # remove variables following the fixes request
        data = self.operator.delete_variables(data)

        # Fix coordinates according to a given data model
        data = self.datamodel.apply_datamodel(data)


        return data

    def _define_deltat(self, default):
        """
        Define the deltat for the fixer. 
        The priority is given to the metadata, then to the fixes and finally to the default value.
        Return deltat in seconds.
        """

        # First case: get from metadata
        metadata_deltat = self.metadata.get('deltat')
        if metadata_deltat:
            self.logger.debug('deltat = %s read from metadata', metadata_deltat)
            return metadata_deltat

        # Second case if not available: get from fixes
        fix_deltat = self.fixes.get("deltat")
        if fix_deltat:
            self.logger.debug('deltat = %s read from fixes', fix_deltat)
            return fix_deltat
        
        # Third case: get from default
        self.logger.debug('deltat = %s defined as Fixer() default', default)
        return default

 
    
    def _override_tgt_units(self, tgt_units, varfix, var):
        """
        Override destination units for the single variable
        """

        # Override destination units
        fixer_tgt_units = varfix.get("units", None)
        if fixer_tgt_units:
            self.logger.debug('Variable %s: Overriding target units "%s" with "%s"',
                              var, tgt_units, fixer_tgt_units)
            return fixer_tgt_units
        else:
            return tgt_units

    def _override_src_units(self, data, varfix, var, source):
        """
        Override source units for the single variable
        """

        # Override source units
        fixer_src_units = varfix.get("src_units", None)
        if fixer_src_units:
            if "units" in data[source].attrs:
                self.logger.debug('Variable %s: Overriding source units "%s" with "%s"',
                                  var, data[source].units, fixer_src_units)
                data[source].attrs.update({"units": fixer_src_units})
            else:
                self.logger.debug('Variable %s: Setting missing source units to "%s"',
                                  var, fixer_src_units)
                data[source].attrs["units"] = fixer_src_units

        return data

    def _get_variables_grib_attributes(self, var):
        """
        Get grib attributes for a specific variable

        Args:
            vardict: Variables dictionary with fixes
            var: variable name

        Returns:
            Dictionary for attributes following GRIB convention and string with updated variable name
        """
        self.logger.debug("Grib variable %s, looking for attributes", var)
        try:
            attributes = get_eccodes_attr(var, loglevel=self.loglevel).copy()  # The copy is needed because the function in eccodes.py is cached
            shortname = attributes.get("shortName", None)
            self.logger.debug("Grib variable %s, shortname is %s", var, shortname)

            if var not in ['~', shortname]:
                self.logger.debug("For grib variable %s find eccodes shortname %s, replacing it", var, shortname)
                var = shortname

            self.logger.debug("Grib attributes for %s: %s", var, attributes)
        except TypeError:
            self.logger.warning("Cannot get eccodes attributes for %s", var)
            self.logger.warning("Information may be missing in the output file")
            self.logger.warning("Please check your version of eccodes")

        return attributes, var

    def _check_which_variables_to_fix(self, var2fix, destvar):
        """
        Check on which variables fixes should be applied

        Args:
            var2fix: Variables for which fixes are available
            destvar: Variables on which we want to apply fixes

        Returns:
            List of variables on which we want to apply fixes for which fixes are available
        """

        if destvar and var2fix:  # If we have a list of variables to be fixed and fixes are available
            newkeys = list(set(var2fix.keys()) & set(destvar))
            if newkeys:
                var2fix = {key: value for key, value in var2fix.items() if key in newkeys}
                self.logger.info("Variables to be fixed: %s", var2fix)
            else:
                var2fix = None
                self.logger.info("No variables to be fixed")

        return var2fix

    def get_fixer_varname(self, var):
        """
        Load the fixes and check if the variable requested is there

        Args:
            var (str or list):  The variable to be checked

        Return:
            A list of variables to be loaded
        """

        if self.fixes is None:
            self.logger.debug("No fixes available")
            return var

        variables = self.fixes.get("vars", None)
        if variables:
            self.logger.debug("Variables in the fixes: %s", variables)
        else:
            self.logger.warning("Returning the original variable")
            return var

        # double check we have a list
        if isinstance(var, str):
            var = [var]

        # if a source/derived is available in the fixes, replace it
        loadvar = []
        for vvv in var:
            if vvv in variables.keys():

                # get the source ones
                if 'source' in variables[vvv]:
                    loadvar.append(variables[vvv]['source'])

                # get the ones from the equation of the derived ones
                if 'derived' in variables[vvv]:
                    # filter operations
                    required = [s for s in re.split(r'[-+*/]', variables[vvv]['derived']) if s]
                    # filter constants
                    required_strings = [x for x in required if not x.replace('.', '').isnumeric()]
                    if bool(set(required_strings) & set(variables.keys())):
                        self.logger.error("Recursive fixer definition: %s are variables defined in the fixer!",
                                          required_strings)
                        raise KeyError((
                                    "Recursive fixer definition are not supported when selecting variables or working with FDB sources."  # noqa: E501
                                    "Please change the fixes or call the retrieve() without the var arguments")
                                    )

                    loadvar = loadvar + required_strings
            else:
                loadvar.append(vvv)

        self.logger.debug("Variables to be loaded: %s", loadvar)
        return loadvar




"""
Module including string utilities for AQUA
"""

import re
import random
import string

def generate_random_string(length):
    """
    Generate a random string of lowercase and uppercase letters and digits
    """
    letters_and_digits = string.ascii_letters + string.digits
    random_string = ''.join(random.choice(letters_and_digits) for _ in range(length))
    return random_string


def strlist_to_phrase(items: list[str], oxford_comma: bool = False) -> str:
    """
    Convert a list of str to a english-consistent list.
       ['A'] will return "A"
       ['A','B'] will return "A and B"
       ['A','B','C'] will return "A, B and C" (oxford_comma=False)
       ['A','B','C'] will return "A, B, and C" (oxford_comma=True)
       
    Args:
        items (list[str]): The list of strings to format.
    """
    if not items: return ""
    if len(items) == 1: return items[0]
    if len(items) == 2: return f"{items[0]} and {items[1]}"
    return ", ".join(items[:-1]) + (", and " if oxford_comma else " and ") + items[-1]


def lat_to_phrase(lat: int) -> str:
    """
    Convert a latitude value into a string representation.

    Returns:
        str: formatted as "<deg>°N" for northern latitudes or "<deg>°S" for southern latitudes.
    """
    if lat >= 0:
        return f"{lat}°N"
    if lat < 0:
        return f"{abs(lat)}°S"


def get_quarter_anchor_month(freq_string: str) -> str:
    """
    Get the anchor month from a quarterly frequency string.
    Examples: 'QE-DEC' -> 'DEC'; 'Q-DEC' -> 'DEC'; 'QS' -> 'DEC' (default)
    Args:
        freq_string (str): The frequency string to extract the anchor month from.
    Returns:
        str: The anchor month.
    """
    if '-' in freq_string:
        return freq_string.split('-')[1]
    return 'DEC'


def clean_filename(filename: str) -> str:
    """
    Check a filename by replacing spaces with '_' and forcing lowercase.
    
    Args:
        filename (str): The filename (or part of filename) to check.
        
    Returns:
        str: Filename with spaces replaced by '_' and forced lowercase.
    """
    return filename.replace(' ', '_').lower()


def extract_literal_and_numeric(text):
    """
    Given a string, extract its literal and numeric part
    """
    # Using regular expression to find alphabetical characters and digits in the text
    match = re.search(r'(\d*)([A-Za-z]+)', text)

    if match:
        # If a match is found, return the literal and numeric parts
        literal_part = match.group(2)
        numeric_part = match.group(1)
        if not numeric_part:
            numeric_part = 1
        return literal_part, int(numeric_part)
    else:
        # If no match is found, return None or handle it accordingly
        return None, None



"""
Module to identify the nature of coordinates of an Xarray object.
"""
import xarray as xr
import numpy as np
from metpy.units import units

from aqua.core.logger import log_configure

LATITUDE = ["latitude", "lat", "nav_lat"]
LONGITUDE = ["longitude", "lon", "nav_lon"]
TIME = ["time", "valid_time", "forecast_period", "time_counter"]
ISOBARIC = ["plev"]
DEPTH = ["depth", "zlev"]

# Define the target dimensionality (pressure)
pressure_dim = units.pascal.dimensionality
#meter_dim = units.meter.dimensionality

# Function to check if a unit is a pressure unit
def is_isobaric(unit):
    """Check if a unit is a pressure unit."""
    try:
        return units(unit).dimensionality == pressure_dim
    except Exception as e:
        return False

class CoordIdentifier():
    """
    Class to identify the nature of coordinates of an Xarray object.
    It aims at detecting the longitude, latitude, time and any other vertical
    by inspecting the attributes of the coordinates provided by the user.

    Args: 
        coords (xarray.Coordinates): The coordinates of Dataset to be analysed.
        loglevel (str): The log level to use. Default is 'WARNING'.
    """

    def __init__(self, coords: xr.Coordinates, loglevel='WARNING'):
        """
        Constructor of the CoordIdentifier class.
        """
        self.loglevel = loglevel
        self.logger = log_configure(self.loglevel, 'CoordIdentifier')

        if not isinstance(coords, xr.Coordinates):
            raise TypeError("coords must be an Xarray Coordinates object.")
        self.coords = coords

        # internal name definition for the coordinates
        self.coord_dict = {
            "latitude": [],
            "longitude": [],
            "time": [],
            "isobaric": [],
            "depth": []
        }

    def identify_coords(self):
        """
        Identify the coordinates of the Xarray object.
        """

        # define a dictionary with the methods to identify the coordinates
        coord_methods = {
            "latitude": self._identify_latitude,
            "longitude": self._identify_longitude,
            "isobaric": self._identify_isobaric,
            "depth": self._identify_depth,
            "time": self._identify_time,
        }

        # loop on coordinates provided by the user
        for name, coord in self.coords.items():
            self.logger.debug("Identifying coordinate: %s", name)

            # use the methods to detect the coordinates and assign attributes accordingly
            for coord_name, identify_func in coord_methods.items():
                if identify_func(coord):
                    self.logger.debug(coord_name)
                    if coord_name == "time":
                        self.coord_dict[coord_name].append(self._get_time_attributes(coord))
                    else:
                        self.coord_dict[coord_name].append(self._get_attributes(coord, coord_name=coord_name))
                    continue # loop on the next coordinate

        # check if the coordinates are empty or double!
        self.coord_dict = self._clean_coord_dict()

        return self.coord_dict
    
    def _clean_coord_dict(self):
        """
        Clean the coordinate dictionary.
        Set to None the coordinates that are empty.
        If multiple coordinates are found, keep only the first one and log an error
        """
        for key, value in self.coord_dict.items():
            if len(value) == 0:
                self.coord_dict[key] = None
            elif len(value) == 1:
                self.coord_dict[key] = value[0]
            else:
                self.logger.warning("Multiple %s coordinates found: %s. Disabling data model check for this coordinate.",
                                     key, [x['name'] for x in value])
                self.coord_dict[key] = None
        return self.coord_dict
    
    def _get_time_attributes(self, coord):
        """
        Get the attributes of the time coordinates.

        Args:
            coord (xarray.Coordinates): The coordinate to define the attributes.

        Returns:
            dict: A dictionary containing the attributes of the coordinate.
        """
        return {'name': coord.name,
                'dims:': coord.dims,
                'units': coord.attrs.get('units'),
                'calendar': coord.attrs.get('calendar'),
                'bounds': coord.attrs.get('bounds')}
    
    def _get_attributes(self, coord, coord_name="longitude"):
        """
        Get the attributes of the coordinates.

        Args:
            coord (xarray.Coordinates): The coordinate to define the attributes.
            coord_type (str): The type of coordinate ("horizontal" or "vertical").

        Returns:
            dict: A dictionary containing the attributes of the coordinate.
        """
        coord_range = (coord.values.min(), coord.values.max())
        direction = None
        positive = None
        horizontal = ["longitude", "latitude"]
        vertical = ["depth", "isobaric"]

        if coord.ndim == 1 and coord_name in horizontal:
            direction = "increasing" if coord.values[-1] > coord.values[0] else "decreasing"

        if coord_name in vertical:
            positive = coord.attrs.get('positive')
            if not positive:
                if is_isobaric(coord.attrs.get('units')):
                    positive = "down"
                else:
                    positive = "down" if coord.values[0] > 0 else "up"

        attributes = {
            'name': coord.name,
            'dims': coord.dims,
            'units': coord.attrs.get('units'),
            'range': coord_range,
            'bounds': coord.attrs.get('bounds'),
        }

        if coord_name in horizontal:
            attributes['stored_direction'] = direction
        elif coord_name in vertical:
            attributes['positive'] = positive
        
        if coord_name == "longitude":
            attributes['convention'] = self._guess_longitude_range(coord)
            

        return attributes
    
    @staticmethod
    def _guess_longitude_range(longitude) -> str:
        """
        Guess if the longitude range is from 0 to 360 or from -180 to 180,
        ensuring the grid is global.
        """

        # Guess the longitude range
        if np.any(longitude.values < 0):
            return "centered"
        elif np.any(longitude.values > 180):
            return "positive"
        else:
            return "ambigous"

    @staticmethod
    def _identify_latitude(coord):
        """
        Identify the latitude coordinate of the Xarray object.
        """
        if coord.name in LATITUDE:
            return True
        if coord.attrs.get("standard_name") == "latitude":
            return True
        if coord.attrs.get("axis") == "Y":
            return True
        if coord.attrs.get("units") == "degrees_north":
            return True
        return False
    
    @staticmethod
    def _identify_longitude(coord):
        """
        Identify the longitude coordinate of the Xarray object.
        """
        if coord.name in LONGITUDE:
            return True
        if coord.attrs.get("standard_name") == "longitude":
            return True
        if coord.attrs.get("axis") == "X":
            return True
        if coord.attrs.get("units") == "degrees_east":
            return True
        return False
    
    @staticmethod
    def _identify_time(coord):
        """
        Identify the time coordinate of the Xarray object.
        """
        if coord.name in TIME:
            return True
        if coord.attrs.get("axis") == "T":
            return True
        if coord.attrs.get("standard_name") == "time":
            return True
        return False
    
    @staticmethod
    def _identify_isobaric(coord):
        """
        Identify the isobaric coordinate of the Xarray object.
        """
        if coord.name in ISOBARIC:
            return True
        if coord.attrs.get("standard_name") == "air_pressure":
            return True
        if is_isobaric(coord.attrs.get("units")):
            return True
        return False
    
    @staticmethod
    def _identify_depth(coord):
        """
        Identify the depth coordinate of the Xarray object.
        """
        if coord.name in DEPTH:
            return True
        if coord.attrs.get("standard_name") == "depth":
            return True
        if "depth" in coord.name:
            return True
        if "depth" in coord.attrs.get("long_name", ""):
            return True
        return False
    
    

